function [co, I, en, epol]=PHStencil(m, q, points, n, z, tol, mode)
% This is the general routine for polyharmonic (PH) stencils. 
% See the paper for details.
global RBFsign
global RBFfactor
% We need to recover M and d and Q
[M,d]=size(points);
Q=factorial(q-1+d)/(factorial(q-1)*factorial(d)); % dimension of polynomial space
if M<Q
    disp(sprintf('Not enough points, at least M=%d needed for m=%d, q=%d, d=%d',...
        Q, m, q, d));
    return;
end
% Centralize 
points=points-repmat(z,M,1);
% and prepare for scale calculation 
dstsq=sum((points.^2)')'; % total sum of squares, as M-vector over points
% if users want an internal sorting: 
% [dstsort, isort]=sort(dstsq);
% points=points(isort,:);
% 
% now choose a way to scale: 
% hmax=sqrt(dstsq(end)) % max. distance scale 
hmax=sqrt(sum(dstsq)/(M*(d-1))); % RMS scale 
% and perform the scaling:
points=points/hmax;
% and centralize z:
z=zeros(1,d);
%
% Get polynomial values
%
P=polynomials(points, q); %
% and d-variate Laplacian at zero on polynomial basis:
polrhs=getLaplacianOnPolynomials(Q, d);
%
switch (lower(mode))
    case 'greedy'
        % we calculate a close unisolvent subset, like in close mode
        I=mycolumnfinder(P', tol); % finds the first linear independent columns,
        % keeping column order. Users might set a larger tolerance here.
        % Now add the greedy technique
        [co_loc, I]=getgreedystencil(points, I, P, z, q, n, tol);
        co=zeros(M,1);
        % and insert the result
        co(I)=co_loc;
    case 'close'
%         B=[polrhs P';-1 zeros(1,M)];
%         J=mycolumnfinder(B, tol);
%         I=J(2:end);
        %
        % I=mycolumnfinder(P', tol); % finds the first linear independent columns,
        % keeping column order. Users might set a larger tolerance here.
        % rank(P')
        for p=Q:M
            coloc=P(1:p,:)'\polrhs;
            coloc(find(abs(coloc)<tol),1)=0;
            I=find(abs(coloc)>tol);
            res=norm(polrhs-P(I,:)'*coloc(I), Inf);
            if res<tol
                break;
            end
        end
        co=zeros(M,1);
        co(I)=coloc(I);
    case 'kernel'
        % This is the direct kernel solver for an optimal stencil
        A=kermat(points,points)*RBFsign; % kernel matrix
        LP=laplacekermat(points,z)*RBFsign; % Laplacian
        % now solve full MxQ system 
        x=[A P; P' zeros(Q,Q)]\[LP; polrhs];
        % and extract the stencil, setting small values to zero.
        % This only rarely helps to shorten the stencil...
        co=x(1:M,1);
        co(find(abs(co)<tol),1)=0;
        I=find(abs(co)>tol);
    case 'polback' 
        % this is the backslash stencil calculation
        % by directly solving the linear system
        co=P'\polrhs;
        % The following is often unnecessary, because the backslash
        % usually produces exact zeros:
        I=find(abs(co)>tol);
        co(setdiff(1:M,I))=0;
    case 'poll2'
        % This is the pseudoinverse solution
        co=pinv(P')*polrhs;
        I=find(abs(co)>tol);
        co(setdiff(1:M,I))=0;
    case 'poll1'
        % This is the L1 minimal stencil calculation via linear
        % programming. The weights x are split into x=x^+ - x^- for
        % nonnegative x^+ and x^-, whose sum is minimized.
        % Setting up for linprog
        Afull=[P' -P']; % input matrix caring for signs
        p=ones(2*M,1);  % objective vector for L_1
        lb=zeros(2*M,1); % lower bounds
        options = optimset('LargeScale','off','Display','off'); 
        % to guarantee small support, and no success message
        [x, fval, exitflag]=linprog(...
            p, [], [], Afull, polrhs,lb,[],[],options );
        if exitflag<=0
            disp(sprintf('Routine linprog within PHStencil exited with flag %d',exitflag));
            co=zeros(M,1);
            I=[];
            epol=NaN;
            en=NaN;
            return
        end
        % We check for components that are small and set them to zero.
        % This may be risky, but we control the polynomial exactness later
        co=x(1:M,1)-x(M+1:end,1);
        co(find(abs(co)<tol),1)=0;
        I=find(abs(co)>tol);
    otherwise
        disp(sprintf('Mode %s not implemented in PHStencil',mode))
        co=zeros(M,1);
        I=[];
        epol=NaN;
        en=NaN;
        return
end
%
% Error evaluation.
% The stencil is assumed to be supported on points with indices in I.
% We recalculate the data because we want an independent check for ALL modes 
% 
A=kermat(points(I,:),points(I,:))*RBFsign; % kernel matrix
LP=laplacekermat(points(I,:),z)*RBFsign;% Laplacian vector
gamma=laplaceXYkermat(z,z)*RBFsign; % should be zero 
en=sqrt(gamma-2*co(I)'*LP+co(I)'*A*co(I));
epol=norm(P'*co-polrhs,Inf);
%
% scale back
%
en=en*hmax^(m-2-d/2)*sqrt(RBFfactor);
co=co/hmax^2;
% if resorting was made, uncomment this:
% I=isort(I);
return
