function [co, Ico]=getgreedystencil(pts, I, P, z, q, n, tol)
% Calculate stencil for Laplacian at z based on M pts.
% The M times Q array P has the polynomial values.
% The array I holds a list of Q indices of a unisolvent subset.
% The final stencil has coefficients in co and point indices in Ico.
% It has at most n points.
global RBFsign
global RBFfactor
IQ=I; % save the Q indices of unisolvent set for later use
[M, Q]=size(P);
[Mp, d]=size(pts);
% get Lagrange basis on point set 
U=zeros(M, n); % the matrix for the Lagrange/Newton basis
U(:,1:Q)=(P(I,:)'\P')';  % filling the first Q columns 
% tUi=norm(U(I,1:Q)-eye(Q),Inf)
U(I,1:Q)=eye(Q); % bringing roundoff back to zero
% with the polynomial Lagrange basis values we now also define
L=U; % the matrix for the true Lagrange basis
% This finishes the polynomial stuff
%
A=zeros(M,n); % the matrix for kernel values
A(:,1:Q)=kermat(pts,pts(I,:))*RBFsign;   % filling it 
%
PFV=zeros(M,1); % the Power Function Value (squared) vector
for i=1:M  % avoiding an MxM matrix, therefore using a loop
    PFV(i,1)=-2*U(i,1:Q)*A(i,1:Q)'+U(i,1:Q)*A(I,1:Q)*U(i,1:Q)';  
end
% tPFVi=norm(PFV(I),Inf)
PFV(I)=0;  % fixing roundoff 
%
LK=laplacekermat(pts,z)*RBFsign; % the Laplace values of the kernel on pts
RR=getLaplacianOnPolynomials(Q, d);
% transition to the Lagrange basis 
LP=zeros(n,1+n-Q); % we store stencils for Q+i points in LP(:,i+1)
LP(1:Q,1)=P(I,:)'\RR;
% 
% now for the error norm squared
% gamma=laplaceXYkermat(z,z)*RBFsign
% this should be zero if m is large enough, but we might check
% if abs(gamma)>tol
%     disp(sprintf('Double Laplacian of kernel at zero is %g, too large',gamma));
%     disp('Going on with value zero')
% end
% we replace it by exactly zero now
% and use the formulas in the paper. First the quadratic form
gamma=0-2*LP(1:Q,1)'*LK(I,1) +LP(1:Q,1)'*A(I,1:Q)*LP(1:Q,1);
% and then the RLK vector. like in the paper
RLK=LK-A(:,1:Q)*LP(1:Q,1)-U(:,1:Q)*LK(I,1)+U(:,1:Q)*A(I,1:Q)*LP(1:Q,1);
% tRLKI=norm(RLK(I),Inf)
RLK(I)=0; % eliminate roundoff
gammalist=[gamma];
% This ends the stuff for Q points, we now start to be greedy
for i=Q+1:n % iteration
    % disp(sprintf('New greedy step for i = %d',i))
    %
    J=setdiff(1:M,I); % the complementary index set
    quot=RLK(J,1).^2./PFV(J,1); % the quotients we have to check
    [qm, mq]=max(quot);  % take the max and the index in J of the max 
    p=J(mq); % the actual point index
    %
    if PFV(p)<tol
        disp(sprintf('Premature stop because of small value %g of Power Function',...
            PFV(p)))
        RLK_PFV=[RLK PFV]
        step=i
        point_index=p
        break;
    end
    if qm<tol
        disp(sprintf('Premature stop because of small gain %g',qm))
        RLK_PFV=[RLK PFV]
        step=i
        point_index=p
        break;
    end
    denomi=sqrt(PFV(p)); % the denominator of the new Newton basis function 
    % aii=RLK(p)*denomi; % the stencil weight at the new point, a_m^m in the text
    aii=RLK(p)/PFV(p); % the stencil weight at the new point, a_m^m in the text
    gamma=gamma-RLK(p)^2/PFV(p); % the new squared error norm
    if gamma<0
        disp(sprintf('Negative norm %g in step %d', gamma, i));
        RLK_PFV=[RLK PFV]
        point_index=p
        RLKPFVp=[RLK(p) PFV(p)]
        decrement=RLK(p)^2/PFV(p)
        nPFVI=norm(PFV(I),Inf)
        PFVI=PFV(I)
        break;
    end
    % 
    A(:,i)=kermat(pts,pts(p,:))*RBFsign; % get a new column for kernel matrix 
    % and proceed to calculate new Newton basis function. Case Q first
    w=A(:,i)-U(:,1:Q)*A(IQ,i)-A(:,1:Q)*U(p,1:Q)'+U(:,1:Q)*A(IQ,1:Q)*U(p,1:Q)';
    %
    % twIQ=norm(w(IQ),Inf)
    w(IQ)=0; % eliminate roundoff
    % now update for steps after Q
    for j=Q+1:i-1
        w=w-U(:,j)*U(p,j);
    end
    % twI=norm(w(I),Inf)
    w(I)=0; % care for roundoff
    % and put new Newton basis function into U
    U(:,i)=w/denomi;
    % This updates the Power Function 
    PFV=PFV-U(:,i).^2;
    % we update the stencil, using old true Lagrange basis
    LP(i,i-Q+1)=aii;
    LP(1:i-1,i-Q+1)=LP(1:i-1,i-Q)-aii*L(p,1:i-1)';
    % now update the true Lagrange basis 
    L(:,i)=U(:,i)/U(p,i); 
    L(:,1:i-1)=L(:,1:i-1)-L(:,i)*L(p,1:i-1);
    % 
    if abs(w(p))<tol
        disp(sprintf('Premature stop because of small Newton value %g',w(p)))
        RLK_PFV_w=[RLK PFV w]
        break;
    end
    % update RLK
    RLK=RLK-w*RLK(p)/w(p);
    % tRLKI=norm(RLK(I), Inf)
    RLK(I)=0; % care for roundoff
    % update the index set 
    I=[I p];
    % tPFV=norm(PFV(I),Inf)
    PFV(I)=0;
    % testL=norm(L(I,1:i)-eye(i),Inf)
    L(I,1:i)=eye(i); % care for roundoff 
    % tPR=norm(LP(1:i,i-Q+1)'*P(I,:)-RR',Inf) % check exactness on polys
    gammalist=[gammalist gamma];
    %
    % additional check on stencil
%     Ac=kermat(pts(I,:),pts(I,:))*RBFsign;
%     Lc=laplacekermat(pts(I,:),z)*RBFsign;
%     Pc=P(I,:);
%     cc=[Ac Pc; Pc' zeros(Q,Q)]\[Lc; RR];
%     coc=cc(1:i,1)
%     LP(1:i,1) 
%     quotsc=coc./LP(1:i,1) 
%     testnsc=sqrt((-2*Lc'*coc+coc'*Ac*coc)*RBFfactor)
end
% norms=sqrt([gammalist(1) gamma]*RBFfactor)
% if one wants to see how the error decreases, uncomment the following:
% figure
% plot(Q:Q+length(gammalist)-1,sqrt(gammalist*RBFfactor))
% title(sprintf('Error norm for M=%d. n=%d, Q=%d, q=%d', M, n, Q, q))
% xlabel('Number of used points')
% startnorm=sqrt(gammalist(1)*RBFfactor);
% endnorm=sqrt(gammalist(end)*RBFfactor);
%
% now check for a smart subset 
ip=find(sqrt(gammalist)<sqrt(gammalist(end))*1.5);
ipfirst=ip(1);
Ico=I(1:Q+ipfirst-1);
co=LP(1:Q+ipfirst-1,ipfirst);
% use the following if there is no subset selection 
% Ico=I;
% co=LP(1:n,length(gammalist));
